import numpy as np
import math

def todo_frames(signal, framelength, frameskip):
    '''
    Chop the waveform into frames
    self.frames[t,n] should be self.signal[t*self.frameskip+n]
    '''
    nframes = 1+int(math.ceil((len(signal)-framelength)/frameskip))
    frames = np.zeros((nframes,framelength))
    raise NotImplementedError('You need to write this part!')
    return(frames)
                                                     
def todo_autocor(frames):
    '''
    Find the autocorrelation function of each frame
    autocor[t,m] should equal R[m], where R[m] = sum_n(frame[n] frame[n-m]).
    '''
    (nframes, framelength) = frames.shape
    autocor = np.zeros((nframes,framelength))
    raise NotImplementedError('You need to write this part!')
    return(autocor)
        
def todo_lpc(autocor, order):
    '''
    Calculate the LPC coefficients in each frame
    Inputs:
    autocor (nframes,framelength) - autocorrelation coefficients of each frame
    order (scalar) - order of the LPC analysis
    Outputs:
    lpc (nframes,order) - LPC predictor coefficients in each frame
    gamma (nframes,order) - vector of R[1] through R[order] for each frame
    Rmatrix (nframes,order,order) - for each frame, a matrix of R[0] through R[order-1]
    '''
    (nframes, framelength) = autocor.shape
    lpc = np.zeros((nframes,order))
    gamma = np.zeros((nframes,order))
    Rmatrix = np.zeros((nframes,order,order))
    raise NotImplementedError('You need to write this part!')
    return(lpc,gamma,Rmatrix)
            
def todo_framepitch(autocor, samplerate):
    '''
    Calculate the pitch period in each frame:
    pitch[t] = 0 if the frame is unvoiced
    pitch[t] = pitch period, in samples, if the frame is voiced.
    Pitch period should maximize R[pitch]/R[0], in the range ceil(0.04*Fs) <= pitch < floor(0.013*Fs)
    Call the frame voiced if and only if R[pitch]/R[0] >= 0.25.
    '''
    (nframes,framelength)=autocor.shape
    minpitch = int(math.ceil(0.004*samplerate))
    maxpitch = int(math.floor(0.013*samplerate))
    framepitch = np.zeros(nframes)
    raise NotImplementedError('You need to write this part!')
    return(framepitch)
            
def todo_framelevel(frames):
    '''
    Calculate the level (energy in decibels) of each frame
    level[t] = 20*log10(sqrt(average(square(frame[t,:]))))
    '''
    (nframes,framelength) = frames.shape
    framelevel = np.zeros((nframes))
    raise NotImplementedError('You need to write this part!')
    return(framelevel)

def todo_samplelevel(framelevel, frameskip):
    '''
    Linearly interpolate framelevel, between frame boundaries,
    in order to find the level (log standard deviation) of each output sample.
    samplelevel[t,n] = framelevel[t]*(frameskip-n)/frameskip + framelevel[t+1]*n/frameskip
    '''
    nframes = len(framelevel)
    samplelevel = np.zeros((nframes-1,frameskip))
    upramp = np.array([ n/frameskip for n in range(0,frameskip) ])
    raise NotImplementedError('You need to write this part!')
    return(samplelevel)
        
def todo_samplepitch(framepitch, frameskip):
    '''
    Linearly interpolate pitch, between frame boundaries,
    If t and t+1 voiced:
      samplepitch[t,n] = framepitch[t]*(frameskip-n)/frameskip + framepitch[t+1]*n/frameskip
    If only t voiced: samplepitch[t,n] = framepitch[t]
    If t unvoiced: samplepitch[t,n] = 0
    '''
    nframes = len(framepitch)
    samplepitch = np.zeros((nframes-1,frameskip))
    upramp = np.array([ n/frameskip for n in range(0,frameskip) ])
    raise NotImplementedError('You need to write this part!')
    return(samplepitch)
                
def todo_excitation(samplepitch):
    '''
    Inputs:
    samplepitch (nsynthframes,frameskip) - effective pitch period for every output sample
    Outputs:
    phase (nsynthframes,frameskip) - phase of the excitation, for every output sample.
      equal to zero if the frame is unvoiced.
      If the frame is voiced, phase[t,n] = phase[t,n-1] + 2pi/samplepitch[t,n].
    excitation (nsynthframes,frameskip) - the excitation signal
      Equal to np.random.normal if the frame is unvoiced.
      Otherwise, equal to an impulse whenever phase passes an integer multiple of 2pi.
      The magnitude of the pitch pulse is sqrt(samplepitch), to make sure RMS=1.
    '''
    (nsynthframes, frameskip) = samplepitch.shape
    np.random.seed(0)
    excitation = np.zeros((nsynthframes,frameskip))
    phase = np.zeros((nsynthframes,frameskip))
    raise NotImplementedError('You need to write this part!')
    return(phase,excitation)

def todo_stable(lpc):
    '''
    Inputs:
    lpc (nframes,order): LPC predictor coefficients
    Outputs:
    lpcpoly (nframes,order+1): coefficients of A(z) in each frame
    lpcroots (nframes,order): roots of A(z) in each frame
    stableroots (nframes,order): roots of A(z), clipped to a max amplitude of 0.98
    stablepoly (nframes,order+1): coefficients of the stabilified A(z)
    stablecofs (nframes,order): predictor coefficients (negative of the polynomial)
    Hint:
    Read the documentation about np.poly and np.roots.
    '''
    (nframes,order)=lpc.shape
    lpcpoly = np.zeros((nframes,order+1))
    lpcroots = np.zeros((nframes,order),dtype=complex)
    stableroots = np.zeros((nframes,order),dtype=complex)
    stablepoly = np.zeros((nframes,order+1))
    stablecofs = np.zeros((nframes,order))
    raise NotImplementedError('You need to write this part!')
    return(lpcpoly, lpcroots, stableroots, stablepoly, stablecofs)
            
def todo_synthesis(excitation, stablecofs, samplelevel):
    '''
    Synthesize the speech.
      First, reshape the excitation into a single vector.
      Second, scale it by np.power(10, samplelevel/20).
      Third, use stablecofs to predict the speech sample, and add it to each excitation.
      Be sure to always use the stable cofs from the right frame!
    '''
    (nsynthframes, frameskip) = excitation.shape
    (nframes, order) = stablecofs.shape
    synthesis = np.zeros(nsynthframes*frameskip)
    raise NotImplementedError('You need to write this part!')
    return(synthesis)

#################################################################
# What to do if somebody calls this from the command line:
if __name__=="__main__":
    import argparse,h5py,wave
    parser = argparse.ArgumentParser(description='''Run MP5 using sample data, and save results.''')
    parser.add_argument('-w','--wav_filename',default='data/file0.wav',
                        help='''wav_filename.  Default: data/file0.wav''')
    args = parser.parse_args()

    with wave.open(args.wav_filename,'rb') as w:
        samplerate = w.getframerate()
        nsamps = w.getnframes()
        signal = np.frombuffer(w.readframes(nsamps),dtype=np.int16).astype('float32')
    framelength = round(0.03*samplerate)
    frameskip = round(0.015*samplerate)
    order = 10

    frames = todo_frames(signal, framelength, frameskip)
    autocor = todo_autocor(frames)
    lpc, gamma, Rmatrix = todo_lpc(autocor, order)
    framepitch = todo_framepitch(autocor, samplerate)
    framelevel = todo_framelevel(frames)
    samplelevel = todo_samplelevel(framelevel, frameskip)
    samplepitch = todo_samplepitch(framepitch, frameskip)
    lpcpoly, lpcroots, stablepoly, stableroots, stablecofs = todo_stable(lpc)
    phase,excitation = todo_excitation(samplepitch)
    synthesis = todo_synthesis(excitation, stablecofs, samplelevel)
    with wave.open('result.wav','wb') as w:
        maxsamp = np.amax(np.abs(synthesis))
        y = np.round(32768*synthesis/maxsamp).astype('int16')
        w.setnchannels(1)
        w.setsampwidth(2)
        w.setframerate(samplerate)
        w.setnframes(len(y))
        w.writeframesraw(y)

    with h5py.File('results.hdf5','w') as f:
        f.create_dataset('signal', data=signal)
        f.create_dataset('frames', data=frames)
        f.create_dataset('autocor',data=autocor)
        f.create_dataset('lpc',data=lpc)
        f.create_dataset('gamma',data=gamma)
        f.create_dataset('Rmatrix',data=Rmatrix)
        f.create_dataset('lpcpoly',data=lpcpoly)
        f.create_dataset('lpcroots',data=lpcroots)
        f.create_dataset('stablepoly',data=stablepoly)
        f.create_dataset('stableroots',data=stableroots)
        f.create_dataset('stablecofs',data=stablecofs)
        f.create_dataset('framepitch',data=framepitch)
        f.create_dataset('framelevel',data=framelevel)
        f.create_dataset('samplelevel',data=samplelevel)
        f.create_dataset('samplepitch',data=samplepitch)
        f.create_dataset('phase',data=phase)
        f.create_dataset('excitation',data=excitation)
        f.create_dataset('synthesis',data=synthesis)

    print('mp5 completed successfully; results saved in results.hdf5')
    print('Now try running mp5overview.ipynb, or run_tests.py')
    
