import unittest, h5py, mp5
from gradescope_utils.autograder_utils.decorators import weight
import numpy as np

# TestSequence
class TestStep(unittest.TestCase):
    def setUp(self):
        self.h5 = h5py.File('solutions.hdf5','r')
        self.samplerate = 11025
        self.frameskip = round(0.015*self.samplerate)

    @weight(5.56)
    def test_frames(self):
        framelength = round(0.03*self.samplerate)
        frames = mp5.todo_frames(self.h5['signal'], framelength, self.frameskip)
        e = np.sum(np.abs(frames-self.h5['frames']))
        self.assertTrue(e < 0.04, 'todo_frames wrong by %d%% (visible case)'%(int(100*e)))

    @weight(5.55)
    def test_autocor(self):        
        autocor = mp5.todo_autocor(self.h5['frames'])
        e = np.sum(np.abs(autocor-self.h5['autocor']))
        self.assertTrue(e < 0.04, 'todo_autocor wrong by %d%% (visible case)'%(int(100*e)))
        
    @weight(5.56)
    def test_lpc(self):        
        lpc, gamma, Rmatrix = mp5.todo_lpc(self.h5['autocor'], 10)
        e = np.sum(np.abs(lpc-self.h5['lpc']))
        self.assertTrue(e < 0.04, 'todo_lpc wrong by %d%% (visible case)'%(int(100*e)))
        
    @weight(5.55)
    def test_framepitch(self):        
        framepitch = mp5.todo_framepitch(self.h5['autocor'], self.samplerate)
        e = np.sum(np.abs(framepitch-self.h5['framepitch']))
        self.assertTrue(e < 0.04, 'todo_framepitch wrong by %d%% (visible case)'%(int(100*e)))

    @weight(5.56)
    def test_framelevel(self):        
        framelevel = mp5.todo_framelevel(self.h5['frames'])
        e = np.sum(np.abs(framelevel-self.h5['framelevel']))
        self.assertTrue(e < 0.04, 'todo_framelevel wrong by %d%% (visible case)'%(int(100*e)))

    @weight(5.55)
    def test_samplelevel(self):                
        samplelevel = mp5.todo_samplelevel(self.h5['framelevel'], self.frameskip)
        e = np.sum(np.abs(samplelevel-self.h5['samplelevel']))
        self.assertTrue(e < 0.04, 'todo_samplelevel wrong by %d%% (visible case)'%(int(100*e)))
        
    @weight(5.56)
    def test_samplepitch(self):                
        samplepitch =  mp5.todo_samplepitch(self.h5['framepitch'], self.frameskip)
        e = np.sum(np.abs(samplepitch-self.h5['samplepitch']))
        self.assertTrue(e < 0.04, 'todo_samplepitch wrong by %d%% (visible case)'%(int(100*e)))
        
    @weight(5.55)
    def test_stable(self):                
        lpcpoly, lpcroots, stablepoly, stableroots, stablecofs = mp5.todo_stable(self.h5['lpc'])
        e = np.sum(np.abs(stablecofs-self.h5['stablecofs']))
        self.assertTrue(e < 0.04, 'todo_stable wrong by %d%% (visible case)'%(int(100*e)))

    @weight(5.56)
    def test_excitation(self):                
        phase,excitation = mp5.todo_excitation(self.h5['samplepitch'])
        e = np.sum(np.abs(phase-self.h5['phase']))
        self.assertTrue(e < 0.04, 'todo_excitation phase wrong by %d%% (visible case)'%(int(100*e)))
        
